\name{mapSignatureSets}
\alias{mapSignatureSets}
\title{
  mapSignatureSets
}
\description{
  Find a mapping from one set of signatures to another. Both Alexandrov
  and Shiraishi signatures can be handled, but both sets must be of the
  same type. The mapping can either be a unique (one-to-one) mapping or
  indentify best matching while allowing multiple signatures to be
  mapped to the same target signature if it is the best match for more
  than one signature.
}
\usage{
mapSignatureSets(fromSignatures, toSignatures, method="euclidean",
                 unique=FALSE)
} 
\arguments{
  \item{fromSignatures}{
    (Mandatory) A set (list) of signatures of the Alexandrov (vector) or
    Shiraishi type (data frame or matrix), that has to be mapped to the
    signatures of a second set (\code{toSignatures}). 
  }
  \item{toSignatures}{
    (Mandatory) The set (list) of signatures to which the set of
    \code{fromSignatures} has to be mapped.
  }
  \item{method}{
    (Optional) The distance measure to be used. This can be one of the
    following: "\code{frobenius}" for Frobenius distance between
    matrices (only for Shiraishi signatures); "\code{rss}" for the
    residual sum of squares (squared error); or any distance measure
    available for the function \code{dist()} of the \code{stats}
    package. Default: "\code{euclidean}".
  }
  \item{unique}{
    (Optional) If set to \code{FALSE} (default), then for each signature
    of \code{fromSignatures} the best match (minimum distance) from
    \code{toSignatures} is selected. The selected signatures need not be
    unique, i.e., one signature of \code{toSignatures} may be the best
    match for multiple signatures of \code{fromSignatures}. If set to
    \code{TRUE}, i.e., if a unique (one-to-one) mapping is required, an
    interative approach is performed: in each step, the best matching
    pair from \code{fromSignatures} and \code{toSignatures} is mapped
    and then removed from the list of signatures that remain to be
    mapped, such that they cannot be selected again.
  }
}
%\details{
%}
\value{
  A vector having as elements the mapped signatures of
  \code{fromSignatures}, and as names the signatures of
  \code{toSignatures} with which they have been associated.
}
\references{
Krueger, Piro (2017) Identification of Mutational Signatures Active in
Individual Tumors. NETTAB 2017 - Methods, Tools & Platforms for
Personalized Medicine in the Big Data Era, October 16-18, 2017, Palermo,
Italy. PeerJ Preprints 5:e3257v1, 2017.\cr
http://rmpiro.net/decompTumor2Sig/
}
\author{
  Rosario M. Piro\cr
  Freie Universitaet Berlin\cr
  Maintainer: Rosario M. Piro\cr
  E-Mail: <r.piro@fu-berlin.de> or <rmpiro@gmail.com>
}
%\note{
%}


\seealso{
  decompTumor2Sig\cr
  determineSignatureDistances\cr
}
\examples{
### get Alexandrov signatures from COSMIC
signAlexandrov <- loadAlexandrovSignatures()

### convert them to Shiraishi signatures
signAlex2Shi <- convertAlexandrov2Shiraishi(signAlexandrov)

### define a small set of arbitrary signatures just for testing
### (similar to signatures 1, 5 and 13, respectively)
test1 <- matrix(c(0.1,   0, 0.7, 0.1, 0.1,   0,
                  0.3, 0.2, 0.3, 0.2,   0,   0,
                  0.2, 0.1, 0.5, 0.2,   0,   0), nrow=3, byrow=TRUE) 
test2 <- matrix(c(0.1, 0.1, 0.3, 0.1, 0.3, 0.1,
                  0.3, 0.25, 0.2, 0.25, 0,   0,
                  0.3, 0.2, 0.2, 0.3,   0,   0), nrow=3, byrow=TRUE)
test3 <- matrix(c(0.1, 0.7, 0.2,   0,   0,   0,
                    0,   0,   0, 1.0,   0,   0,
                  0.5, 0.1,   0, 0.4,   0,   0), nrow=3, byrow=TRUE)
fromSig <- list(sig1=test1, sig2=test2, sig3=test3)

### compute distances of the test signature to the converted
### Alexandrov signatures from COSMIC
mapSignatureSets(fromSig, signAlex2Shi, method="frobenius", unique=TRUE)
}
\keyword{mutational signatures}
\keyword{mutation signatures}
\keyword{mutations}
\keyword{tumor genomes}
